"""
Various utilities used internally.
"""

from __future__ import annotations

from collections import namedtuple
from collections.abc import Callable, Iterable, Mapping
from math import sqrt
import re
from typing import Any, Optional, TypeVar, Union


RGB = namedtuple("RGB", "r g b")

T = TypeVar("T")

T_RGB = Union[tuple[int, int, int], RGB]
T_HLS = tuple[float, float, float]

HEX = Union[str, int]


def _copydoc(
    source: Any,
    replace: Optional[
        Mapping[
            re.Pattern[str],
            Union[
                str,
                Callable[[re.Match[str]], str]
            ]
        ]
    ] = None,
) -> Callable[[T], T]:
    """
    Copies docstring of another function.
    
    Parameters
    ----------
    source
        The object whose text should be copied.
    
    replace
        A mapping of regex patterns and either callables or
        string to replace the matches with.
    """
    def decorator(fn: T) -> T:
        doc = source.__doc__
        
        if replace is not None:
            for pattern, replace_with in replace.items():
                doc = pattern.sub(
                    replace_with,
                    doc
                )
        
        fn.__doc__ = doc
        return fn
    return decorator

def _get_closest_color(
    color: RGB,
    colors: Iterable[tuple[T, RGB]]
) -> T:
    """
    Returns first value of a tuple by finding
    the closest rgb color.
    
    Parameters
    ----------
    color
        The rgb color to find the closest possible
        rgb color from the ``colors`` parameter.
    
    colors
        Sequence of two-sized tuples. Second value is
        an rgb value that ``color`` could approach.
        The first value is any value that will be returned
        if the second value is the closest to ``color``.
    
    Returns
    -------
    Associated value of closest color.
    """
    diffs = {}
    for value, rgb in colors:
        if color == rgb:
            return value
        
        diff = sqrt(
            abs(color.r - rgb.r) ** 2 +
            abs(color.g - rgb.g) ** 2 +
            abs(color.b - rgb.b) ** 2
        )
        
        diffs[diff] = value
    
    return diffs[min(diffs.keys())]




"""
.. versionadded:: 0.1.0

Load ansi objects from a stylesheet file in
your project.
"""

# TODO: What if another library imports a library
#       that has a stylesheet? Maybe add check for
#       a package or some kind of labeling.

from __future__ import annotations

from functools import cache
from pathlib import Path
from types import ModuleType
from typing import Union

from .ansi import Ansi


_globals: dict[str, Ansi] = dict()
"""Loaded style objects."""

_loaded: set[ModuleType] = set()
"""Loaded stylesheet modules."""


def remove_style(name: str) -> None:
    """
    Removes a style.
    
    Parameters
    ----------
    name
        The name of the style object.
    """
    del _globals[name]

def remove_all() -> None:
    """
    Removes all styles.
    """
    _globals.clear()

def export(**styles: Ansi) -> None:
    """
    Loads all objects.
    
    This usually called from a stylesheet module
    in order to make the ansi objects globally
    available.
    
    .. admonition:: Preference
       :class: tip
       
       This method is preferred over :func:`load_stylesheet`.
    
    .. caution::
       
       This will not filter out ansi objects but
       rather raise a ``TypeError``.
    
    Parameters
    ----------
    styles
        Ansi objects and an associated name.
    """
    for name, ansi in styles.items():
        if not isinstance(ansi, Ansi):
            raise TypeError(
                f"{name!r} ({ansi!r}) is not an "
                f"instance of {Ansi!r}"
            )
        _globals[name] = ansi

@cache
def load_stylesheet(
    source: Union[Path, ModuleType, str, None] = None,
    force: bool = False
) -> None:
    """
    Loads a stylesheet.
    
    .. admonition:: Preference
       :class: tip
       
       :func:`export` is preferred over this method.
    
    
    .. tip::
       
       The safest method is to import the stylesheet
       module and use it as the ``source`` argument.
       
       Alternatively the :func:`export` function may
       be used which is also safe.
    
    Parameters
    ----------
    source
        When a ``Path`` or ``str`` is given, it
        will load all style objects from that file
        by importing it. If the path is a directory
        it will try to find a file named ``_stylesheet.py``
        or ``stylesheet.py`` (in that order) and use
        it as a stylesheet.
        
        This function will attempt to load all style
        objects mentioned in the ``__all__`` variable.
        If that variable is not defined, it will filter
        all style objects defined in the module.
        
        .. caution::
           
           This will include built-in style objects such
           as ones imported from :mod:`adorable.style`.
        
        When no source is provided (default) the function
        will attempt to search the stylesheet file by
        querying the module that called this function.
        Then it will use the directory
        where the caller's source file is as the path. After that
        the same process takes place as when defining source
        with a ``Path``.
    
    force
        This function caches already loaded stylesheets -
        no matter what value the ``source`` parameter
        previously had. Forcing a rerun may be done by
        setting this to ``True``.
    
    Raises
    ------
    ``RuntimeError``
        Could not find source file.
    """
    import importlib.util as imp
    import inspect
    import sys
    
    from . import filter_ansi
    
    if source is None:
        callers = inspect.stack()
        for caller in reversed(callers):
            filename = caller.filename
            
            if filename == "<string>":
                continue
            
            source = Path(filename)
            break
        
        else:
            raise RuntimeError("cannot find source file")
        
        parent = source.parent
        
        for match in ["_stylesheet.py", "stylesheet.py"]:
            source = parent / match
            if source.exists():
                break
        
        else:
            raise RuntimeError("cannot find stylesheet")
    
    if isinstance(source, str):
        source = Path(source)
    
    if isinstance(source, Path):
        spec = imp.spec_from_file_location("_stylesheet_", str(source))
        if spec is None:
            raise RuntimeError("cannot load stylesheet")
        
        source = imp.module_from_spec(spec)
        if spec.loader is None:
            raise RuntimeError("cannot load stylesheet")
        
        sys.modules[source.__name__] = source
        spec.loader.exec_module(source)
    
    try:
        all_ = source.__all__
    except AttributeError:
        objects = source.__dict__
    else:
        objects = dict()
        for name in all_:
            objects[name] = source.__dict__[name]
    
    export(**filter_ansi(objects))
    _loaded.add(source)




"""
Common ansi styles.
"""

from __future__ import annotations
from typing import Any
from .ansi import Ansi

class Style(Ansi):
    def __init__(self, enable: Any, disable: Any):
        self._ansi = [enable]
        self._off = [disable]


BOLD = Style(1, 22)           # b
"""Bold style"""

DIM = Style(2, 22)            # d
"""Dim style"""

ITALIC = Style(3, 23)         # i
"""Italic style"""

UNDERLINE = Style(4, 24)      # u
"""Underline style"""

BLINK = Style(5, 25)          # f (flash)
"""Blink style"""

INVERSE = Style(7, 27)        # r (reverse)
"""Inverse style"""

INVISIBLE = Style(8, 28)      # h (hidden)
"""Invisible style"""

STRIKETHROUGH = Style(9, 29)  # s
"""Strike through style"""



"""
Utilities for creating color objects.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from copy import copy
from enum import auto, IntEnum
import operator
from typing import Any, Union
import warnings

from . import _palette
from . import webcolors
from .ansi import Ansi, _get_ansi_string, paint
from .term import Terminal
from .utils import _copydoc, _get_closest_color,  RGB, HEX, T_RGB


_NOT_INITIALIZED: str = "color not initialized"
_ALREADY_INITIALIZED: str = "color already initialized"


class _GroundError(Exception):
    pass


class _Ground(IntEnum):
    NONE = auto()
    FORE = auto()
    BACK = auto()
    BOTH = auto()
    
    @staticmethod
    def check(
        *checks: tuple[
            Callable[..., bool],
            Union[Color, int],
            Union[Color, int],
            str
        ]
    ) -> None:
        """
        Parameters
        ----------
        checks
            Each entry contains four fields.
            
            # A callable (usually a function from the
              :pylib:`operator` library).
            # A :class:`_Ground` or :class:`Color` instance.
              (If latter is given, the ground is extracted
              from it.)
            # Another one for comparison.
            # A string to display when the comparison
              returns ``False``.
        
        Raises
        ------
        ``TypeError``
            Entry contains value with invalid type.
        
        ``_GroundError``
            The comparison returned ``False``.
        """
        msg = "invalid ground set"  # in case a string is missing
        for check in checks:
            args: list[int] = []
            for arg in check[1:]:
                if isinstance(arg, Color):
                    args.append(arg._ground)
                
                elif isinstance(arg, str):
                    msg = arg
                
                elif isinstance(arg, int):
                    args.append(arg)
                
                else:
                    raise TypeError(
                        "expected 'Color', 'int' or 'str', "
                        f"got {arg.__class__.__name__!r}"
                    )
            
            if not check[0](*args):
                raise _GroundError(msg)

class Color(Ansi, metaclass = ABCMeta):
    """
    Abstract Base Class for terminal dependent
    colors.
    """
    _termtype: Terminal
    
    def __init__(self, **data: Any):
        """
        .. attention::
           
           Directly initializing this class is
           prohibited. Use one of the available
           class methods such as
           :meth:`from_name`, :meth:`from_rgb` or
           :meth:`from_hex`.
        
        Parameters
        ----------
        data
            Various metadata set by subclasses
            for internal use.
        """
        self._ansi: list[Any] = []
        self._off = [0]
        self._data = data
        self._ground: _Ground = _Ground.NONE
    
    def __str__(self) -> str:
        if not self.is_initialized():
            warnings.warn(
                f"{self.__class__.__name__} instance at "
                f"{hex(id(self))} is not initialized. Use "
                "`.bg`, `.fg` or `.on()` to initialize it.",
                RuntimeWarning
            )
            return ""
        
        return self.enable_str()
    
    def __or__(self, other: Color) -> Color:
        """
        Returns the color that is supported. Returns
        the color with a larger color palette when
        both are supported. Returns the color with a
        smaller color palette when neither is supported.
        """
        if not isinstance(other, Color):
            return NotImplemented
        
        if other.is_supported() and self.is_supported():
            return max(self, other, key = lambda x: x._termtype)
        
        if other.is_supported():
            return other
        
        if self.is_supported():
            return self
        
        return min(self, other, key = lambda x: x._termtype)
    
    def __add__(self, other: Ansi) -> Ansi:
        if isinstance(other, Color):
            _Ground.check((
                operator.ne,
                other, _Ground.NONE,
                _NOT_INITIALIZED
            ), (
                operator.ne,
                other, self,
                "both colors are set to the same ground"
            ))
        
        elif not isinstance(other, Ansi):
            return NotImplemented
        
        obj = copy(self)
        return obj.extend(other)
    
    def __iadd__(self, other: Ansi) -> Ansi:
        if isinstance(other, Color):
            _Ground.check((
                operator.ne,
                other, _Ground.NONE,
                _NOT_INITIALIZED
            ), (
                operator.ne,
                other, self,
                "both colors are set to the same ground"
            ))
        
        elif not isinstance(other, Ansi):
            return NotImplemented
        
        return self.extend(other)
    
    def __call__(self, *args: Any, **kwargs: Any) -> str:
        _Ground.check((
            operator.ne,
            self, _Ground.NONE,
            _NOT_INITIALIZED
        ))
        
        return paint(*args, style = self, **kwargs)
    
    def enable_str(self) -> str:
        _Ground.check((
            operator.ne,
            self, _Ground.NONE,
            _NOT_INITIALIZED
        ))
        
        return super().enable_str()
    
    def disable_str(self) -> str:
        _Ground.check((
            operator.ne,
            self, _Ground.NONE,
            _NOT_INITIALIZED
        ))
        
        return super().disable_str()
    
    def is_initialized(self) -> bool:
        """
        .. versionadded:: 0.1.0
        
        Checks if the color is initialized or not.
        """
        return self._ground != _Ground.NONE
    
    def is_supported(self) -> bool:
        """
        .. versionadded:: 0.1.1
        
        Checks if the color system is supported
        or not.
        """
        return self._termtype.is_supported()
    
    def _get_ansi(self) -> str:
        return _get_ansi_string(self._ansi)
    
    @abstractmethod
    def _form(self, ground: _Ground) -> list[int]:
        """
        Forms the ansi sequence by returning a list
        of its attributes.
        """
        ...
    
    @property
    def fg(self) -> Color:
        """
        Sets the color to foreground mode.
        """
        _Ground.check((
            operator.eq,
            self, _Ground.NONE,
            _ALREADY_INITIALIZED
        ))
        
        self._ansi.extend(self._form(_Ground.FORE))
        self._ground = _Ground.FORE
        
        return self
    
    @property
    def bg(self) -> Color:
        """
        Sets the color to background mode.
        """
        _Ground.check((
            operator.eq,
            self, _Ground.NONE,
            _ALREADY_INITIALIZED
        ))
        
        self._ansi.extend(self._form(_Ground.BACK))
        self._ground = _Ground.BACK
        
        return self
    
    def on(self, other: Color) -> Color:
        """
        Sets the color to foreground mode and adds
        another color as background mode.
        
        Parameters
        ----------
        other
            Background color to set for this style.
            May already be initialized as background
            mode.
        """
        _Ground.check((
            lambda a, b, back, fore: a != back and b != fore,  # type: ignore
            self, other, _Ground.BACK, _Ground.FORE,
            "second color must not be set to background"
        ))
        
        self = self.fg
        self.extend(other.bg if other._ground != _Ground.BACK else other)
        
        self._ground = _Ground.BOTH
        
        return self
    
    @classmethod
    @abstractmethod
    def _from_rgb(cls, rgb: T_RGB) -> Color:
        ...
    
    @classmethod
    def from_rgb(cls, rgb: T_RGB | tuple[float, float, float]) -> Color:
        """
        Get color from RGB value.
        
        .. seealso::
           
           :doc:`Creating Colors <creating-color>`
        """
        values: list[int] = []
        for i in rgb:
            if isinstance(i, float):
                i = round(i * 255)
            values.append(i)
        
        return cls._from_rgb(values)  # type: ignore
    
    @classmethod
    def from_name(cls, name: str) -> Color:
        """
        Get color from Web Color name.
        
        .. seealso::
           
           :doc:`Creating Colors <creating-color>`
        """
        return cls.from_rgb(webcolors.COLORS[name.lower()])
    
    @classmethod
    def from_hex(cls, hex: HEX) -> Color:
        """
        Get color from HEX value.
        
        .. seealso::
           
           :doc:`Creating Colors <creating-color>`
        """
        if isinstance(hex, int):
            # convert to string
            hex = f"{hex:X}"
            
            # leading 0s will be removed so we need
            # to add them (e. g. 0x0AF -> AF -> 0AF)
            if len(hex) < 3:
                hex = hex.zfill(3)
            
            elif 3 < len(hex) < 6:
                hex = hex.zfill(6)
        
        if not isinstance(hex, str):
            raise TypeError("hex value must be int or str")
        
        hex = hex.removeprefix("#")
        
        if len(hex) == 3:
            # convert to a six chars long string
            hex = "".join(char * 2 for char in hex)
        
        elif len(hex) != 6:
            raise ValueError(
                "hex value should consist of 3 or 6 characters "
                "(e. g. `0xFFFFFF` or `0xFFF`)"
            )
        
        h = iter(hex)
        
        rgb = [int(char, 16) * 16 + int(next(h), 16) for char in h]
        return cls.from_rgb(rgb)  # type: ignore

class Color0bit(Color):
    _termtype: Terminal = Terminal.NOCOLOR
    
    def _form(self, ground: _Ground) -> list[int]:
        return []
    
    @classmethod
    def _from_rgb(cls, rgb: T_RGB) -> Color0bit:
        return cls()

class Color3bit(Color):
    _termtype: Terminal = Terminal.BIT3
    
    def _form(self, ground: _Ground) -> list[int]:
        value = self._data["ansi"]
        
        if ground in [_Ground.NONE, _Ground.BOTH]:
            raise ValueError()
        
        return [
            30 + value
        ]
    
    @classmethod
    def _from_rgb(cls, rgb: T_RGB) -> Color3bit:
        color: int = _get_closest_color(
            RGB(*rgb),
            enumerate(_palette.ANSI3BIT)
        )
        return cls(ansi = color, rgb = rgb)

class Color8bit(Color):
    _termtype: Terminal = Terminal.BIT8
    
    def _form(self, ground: _Ground) -> list[int]:
        value = self._data["ansi"]
        
        if ground in [_Ground.NONE, _Ground.BOTH]:
            raise ValueError()
        
        return [
            38 if ground == _Ground.FORE else 48,
            5,
            value
        ]
    
    @classmethod
    def _from_rgb(cls, rgb: T_RGB) -> Color8bit:
        color: int = _get_closest_color(
            RGB(*rgb),
            enumerate(_palette.ANSI8BIT)
        )
        return cls(ansi = color, rgb = rgb)

class Color24bit(Color):
    _termtype: Terminal = Terminal.BIT24
    
    def _form(self, ground: _Ground) -> list[int]:
        r, g, b = self._data["rgb"]
        
        if ground in [_Ground.NONE, _Ground.BOTH]:
            raise ValueError()
        
        return [
            38 if ground == _Ground.FORE else 48,
            2,
            r, g, b
        ]
    
    @classmethod
    def _from_rgb(cls, rgb: T_RGB) -> Color24bit:
        return cls(rgb = rgb)

def get_color() -> type[Color]:
    t = Terminal.get_term([])
    for colortype in [Color0bit, Color3bit, Color8bit, Color24bit]:
        if colortype._termtype == t:
            return colortype
    
    return Color0bit  # unreachable

@_copydoc(Color.from_hex)
def from_hex(*args: Any, **kwargs: Any) -> Color:
    col = get_color()
    return col.from_hex(*args, **kwargs)

@_copydoc(Color.from_rgb)
def from_rgb(*args: Any, **kwargs: Any) -> Color:
    col = get_color()
    return col.from_rgb(*args, **kwargs)

@_copydoc(Color.from_name)
def from_name(*args: Any, **kwargs: Any) -> Color:
    col = get_color()
    return col.from_name(*args, **kwargs)

def empty() -> Color0bit:
    """
    .. versionadded:: 0.1.0
    
    Returns a new instance of :class:`Color0bit`.
    """
    return Color0bit()




"""
Utilities for working with the terminal color
system.
"""

from __future__ import annotations

__all__ = ["Terminal"]

from enum import auto, IntEnum
import os
import sys
from typing import Iterable, Optional, TextIO


cache: Optional[Terminal] = None
"""
Cache for supported terminal color system.
"""

class Terminal(IntEnum):
    """
    .. versionchanged:: 0.1.1
        Removed ``BIT4`` value.
    
    Enumeration for specifying the color system
    to use.
    """
    NOCOLOR = auto()
    BIT3 = auto()
    BIT8 = auto()
    BIT24 = auto()
    
    @classmethod
    def get_term(
        cls,
        stream: Optional[Iterable[TextIO]] = None,
        remember: bool = True,
    ) -> Terminal:
        """
        Returns the color system that is supported
        by the end user's terminal.
        
        The color system is checked in the
        following order:
        
        #. Is ``stream`` a valid terminal?
        #. Environment variable ``ADORABLE_COLOR``
        #. Environment variable ``NO_COLOR``
        #. Environment variable ``COLORTERM``
        #. Environment variable ``TERM``
        #. Fallback: :attr:`NOCOLOR`
        
        .. caution::
            
            The result of this function will be saved in
            the ``cache`` variable. The next time this function
            is called, the result in the cache will be returned
            if it is not ``None``. You can clear the cache at any
            time by setting ``cache`` to ``None``::
                
                from adorable import term
                term.cache = None
            
            .. versionadded:: 0.1.1
        
        Parameters
        ----------
        stream
            The streams to query.
            
            If all of the streams provided are invalid
            terminal, ``RuntimeError`` is raised.
            Defaults to ``stdout`` and ``stderr``.
            
            If you want to skip checking for a
            valid terminal you may set this to an empty
            iterable such as ``[]``.
        
        remember
            .. versionadded:: 0.1.1
            
            Caches the result.
        
        Examples
        --------
        .. code-block::
           
           # check if either stdout or stderr is valid
           Terminal.get_term()
           
           # skip terminal check
           Terminal.get_term([])
           
           # only check stdout
           Terminal.get_term([sys.stdout])
        
        Raises
        ------
        ``RuntimeError``
            Standard output is not a valid terminal.
        """
        global cache
        
        if cache is not None:
            return cache
        
        if stream is None:
            stream = [sys.stdout, sys.stderr]
        
        if any(
            not (st and st.isatty())
            for st in stream
        ):
            raise RuntimeError("standard output is not a valid terminal")
        
        ac = os.getenv("ADORABLE_COLOR")
        if ac == "nocolor":
            res = cls.NOCOLOR
        
        elif ac == "3bit":
            res = cls.BIT3
        
        elif ac == "8bit":
            res = cls.BIT8
        
        elif ac == "24bit":
            res = cls.BIT24
        
        elif int(os.getenv("NO_COLOR", "0")):
            res = cls.NOCOLOR
        
        elif os.getenv("COLORTERM", "0") in ["truecolor", "24bit"]:
            res = cls.BIT24
        
        elif (
            os.getenv("TERM", "")
                .removeprefix("xterm-")
                .removesuffix("color")
        ) == "256":
            res = cls.BIT8
        
        else:
            res = cls.NOCOLOR
        
        if remember:
            cache = res
        return res
    
    def is_supported(self) -> bool:
        """
        .. versionadded:: 0.1.1
        
        Checks if the color system is supported
        by the terminal.
        """
        return self.__class__.get_term() >= self




"""
Markup utilities for adorable.

.. warning::
   
   Using XML for markup is likely going to be replaced in the
   future by a more friendly optimized markup language.

"""

from __future__ import annotations

__all__ = ["insert", "markup_xml"]

import html
import string
from typing import Any, Generator, Optional
import warnings
import xml.etree.ElementTree as ET

from .ansi import Ansi, AnsiNull
from .stylesheet import _globals


class XMLEscapeFormatter(string.Formatter):
    """
    If conversion specifier ``e`` (escape) is
    given or omitted, escape xml contents of
    the value. If ``r`` (raw) is given, insert
    without escaping.
    """
    
    @staticmethod
    def convert_field(value: str, conversion: Optional[str]) -> str:
        if conversion in ["e", None]:
            return html.escape(value)
        
        elif conversion == "r":
            return value
        
        raise ValueError(f"unknown conversion specifier {conversion}")

def get_ansi_from_tag(name: str, style: dict[str, Ansi] = {}) -> Ansi:
    """
    Returns the ansi object in the palette by the
    name. If the ansi object is not present, ``ValueError``
    is raised.
    
    Parameters
    ----------
    name
        The key of the ansi object.
    
    style
        The palette.
    
    Raises
    ------
    ValueError
        Key not present.
    
    Returns
    -------
    The matching ansi object.
    """
    if name in style:
        return style[name]
    
    raise ValueError(f"invalid tag {name!r}")

def style_element(
    element: ET.Element,
    style: dict[str, Ansi] = {},
    _previous_ansi: Ansi = AnsiNull(),
) -> Generator[str, None, None]:
    """
    Yields strings by styling them via their elemt names.
    
    Parameters
    ----------
    element
        The element to style.
    
    style
        The palette.
    
    Other Parameters
    ----------------
    _previous_ansi
        The ansi object used in the last iteration.
    
    Yields
    ------
    Text snippets styled.
    """
    for child in element:
        tag = child.tag
        ansi = get_ansi_from_tag(tag, style = style)
        
        yield ansi.enable_str() + (child.text or "")
        
        yield from style_element(
            child,
            style = style,
            _previous_ansi = ansi
        )
        
        yield (
            ansi.disable_str() +
            _previous_ansi.enable_str() +
            (child.tail or "")
        )

def insert(
    *args: Any,
    **kwargs: Any,
) -> tuple[tuple[Any, ...], dict[str, Any]]:
    """
    Returns ``args`` and ``kwargs`` as is.
    """
    return args, kwargs


def markup_xml(
    string: str,
    style: Optional[dict[str, Ansi]] = None,
    insert: Optional[tuple[tuple[Any, ...], dict[str, Any]]] = None
) -> str:
    """
    .. warning::
       
       Using XML for markup is likely going to be replaced in the
       future by a more friendly optimized markup language.
    
    Use XML to style a string. In order to
    format the string, provide the variables
    directly to this function. All variables
    will be escaped automatically unless ``!r``
    is added at the end. You can also use ``!e``
    to explicitly specify that the variable
    has to be escaped.
    
    Parameters
    ----------
    string
        The markup text.
    
    style
        Mapping of keys and the style to use.
        You may want to use
        :func:`adorable.filter_ansi`.
    
    insert
        Insert variables to string.
        
        .. seealso::
           :func:`insert`
    
    Returns
    -------
    Styled text.
    """
    warnings.warn(
        "XML markup is likely going to be replaced by "
        "another markup language in the future",
        PendingDeprecationWarning,
    )
    
    if style is None:
        style = {}
    
    if insert is None:
        insert = ((), {})
    
    style |= _globals
    text = XMLEscapeFormatter().format(string, *insert[0], **insert[1])
    
    root = ET.fromstring(f"<root>{text}</root>")
    text = root.text or ""
    text += "".join(style_element(root, style = style))
    
    return html.unescape(text)



from __future__ import annotations

from abc import ABC
from copy import copy
import re
from sys import stdout
from typing import Any, Optional, TextIO

from .utils import _copydoc


def paint(*args: Any, style: Optional[Ansi] = None, sep: str = " ") -> str:
    """
    Styles a string.
    
    Parameters
    ----------
    args
        Objects to style. The ``str()`` function will
        be called on each object.
    
    style
        Ansi object that styles the string.
    
    sep
        String that separates ``args``.
    
    Returns
    -------
    The styled string.
    """
    content = sep.join(map(str, args))
    
    if style is None:
        return content
    
    elif not isinstance(style, Ansi):
        raise TypeError(
            "expected `None` or `Ansi` for argument `style`, "
            f"got {style.__class__.__name__}"
        )
    
    return f"{style}{content}{style.disable_str()}"

formatc = paint  # noqa: E305
"""Alias of :func:`paint`."""

def printc(*args: Any, **kwargs: Any) -> None:
    """
    Prints a styled string.
    
    This function takes the same arguments as the
    built-in print function. It also provides
    an extra parameter.
    
    Parameters
    ----------
    args
        Positional arguments passed to :pyfn:`print`
    
    kwargs
        Keyword arguments passed to :pyfn:`print`
    
    style
        Ansi objects that styles the string.
    """
    paint_kwargs = {}
    
    for key in ["style", "sep"]:
        if key in kwargs:
            paint_kwargs[key] = kwargs.pop(key)
    
    print(paint(*args, **paint_kwargs), **kwargs)


class Ansi(ABC):
    _ansi: list[Any]
    _off: list[Any]
    
    def __str__(self) -> str:
        return self.enable_str()
    
    @_copydoc(paint, replace = {
        re.compile("style.+?the string.", re.DOTALL): ""
    })
    def __call__(self, *args: Any, **kwargs: Any) -> str:
        return paint(*args, style = self, **kwargs)
    
    def __add__(self, other: Ansi) -> Ansi:
        """
        Combines two styles.
        """
        obj = copy(self)
        obj.extend(other)
        
        return obj
    
    def __iadd__(self, other: Ansi) -> Ansi:
        """
        Combines this and another style.
        """
        return self.extend(other)
    
    def extend(self, *others: Ansi) -> Ansi:
        """
        Adds other ansi styles to this style.
        """
        for other in others:
            self._ansi.extend(other._ansi)
            self._off.extend(other._off)
        return self
    
    def enable_str(self) -> str:
        """
        Returns escape sequence to enable the ansi style.
        """
        return _get_ansi_string(*self._ansi)
    
    def enable(self, file: Optional[TextIO] = None) -> None:
        """
        Enables the ansi style.
        
        Parameters
        ----------
        file
            File to write to. Defaults to standard output.
        """
        file = file or stdout
        file.write(self.enable_str())
    
    def disable_str(self) -> str:
        """
        Returns escape sequence to disable the ansi style.
        """
        return _get_ansi_string(*self._off)
    
    def disable(self, file: Optional[TextIO] = None) -> None:
        """
        Disables the ansi style.
        
        Parameters
        ----------
        file
            File to write to. Defaults to standard output.
        """
        file = file or stdout
        file.write(self.disable_str())

class AnsiNull(Ansi):
    """
    An implementation of an ansi object that contains
    no data.
    
    Examples
    --------
    A simplified example could look like this::
        
        from typing import Optional
        from adorable.ansi import Ansi, AnsiNull
        
        def apply_ansi(text: str, ansi: Optional[Ansi] = None) -> str:
            if ansi is None:
                ansi = AnsiNull()
            
            return ansi(text)
    
    """
    def __init__(self) -> None:
        self._ansi = []
        self._off = []

def _get_ansi_string(*args: Any) -> str:
    """
    Creates an ansi escape sequence by seperating each
    argument with a semicolon (``;``). Every provided
    argument will be turned into a string.
    
    Examples
    --------
    .. repl::
       
       from adorable.ansi import _get_ansi_string
       _get_ansi_string(38, 5, "20")
    """
    return f"\x1b[{';'.join(map(str, args))}m"




from collections.abc import Mapping
from typing import Any

from .color import (
    Color3bit,
    Color8bit,
    Color24bit,
)

from .style import (
    BOLD,
    DIM,
    ITALIC,
    UNDERLINE,
    BLINK,
    INVERSE,
    INVISIBLE,
    STRIKETHROUGH
)

from .ansi import *
from .markup import *
from .stylesheet import *
from .stylesheet import _globals
from . import term


ANSI_REGEX = re.compile("\x1b\\[.*?[ABCDEFGHJKfnsumlh]")
"""Regex pattern that matches most ansi escape sequences."""

def use(terminal: str) -> None:
    """
    .. versionadded:: 0.2.0b1
    
    A quick way of overriding the current color
    system manually.
    
    Parameters
    ----------
    terminal
        The color system as a string. This might
        be ``"NOCOLOR"``, ``"BIT3"``, ``"BIT8``
        or ``"BIT24"``.
    """
    term.cache = term.Terminal[terminal]

def filter_ansi(style: Mapping[str, Any]) -> dict[str, Ansi]:
    """
    Filters all ansi objects inside a mapping. Useful when
    ansi objects are defined in a module so that ``locals()``
    or ``module.__dict__`` can be used.
    """
    return {k: v for k, v in style.items() if isinstance(v, Ansi)}

